/* 2^n-1 based modular arithmetic.

   Copyright 2001, 2002, 2007 Kevin Ryde

   This file is part of MPM.

   MPM is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2, or (at your option) any later version.

   MPM is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
   FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
   details.

   You should have received a copy of the GNU General Public License along
   with MPM; see the file COPYING.  If not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301
   USA.  */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "gmp.h"
#include "gmp-impl.h"
#include "longlong.h"
#include "mpm.h"
#include "mpm-impl.h"


/* Residues are kept as msize limbs and any value within those limbs, it's
   not not fully reduced to 0<=r<2^n-1, so it's not a unique representation.

   MCARRYSHIFT(m) is the amount to shift a carry out of the high limb when
   adding it back to the low end, ie. msize*GMP_NUMB_BITS - n.  This is zero
   if n is a multiple of GMP_NUMB_BITS, and the code has separate blocks of
   code for carryshift==0 (the "__noshift" routines) or carryshift!=0 (the
   "__shift" routines).

*/


/* change this to "#define TRACE(x) x" for diagnostics */
#define TRACE(x) 

/* re-use fields */
#define MCARRYSHIFT(m) ((m)->_mp_norm_lshift)


static void mpr_2expsub1__neg __GMP_PROTO ((mpr_ptr, mpr_srcptr, mpm_srcptr));


#define ASSERT_MPR_2EXPSUB1(r,m)        \
  do {                                  \
    ASSERT_MPN (RPTR(r), MRSIZE(m));    \
  } while (0)

/* {wp,msize} += c << carryshift, where c is a carry bit, 0 or 1.
   If there's a carry out of that addition, it's added in at the low end the
   same way, ie. newc << carryshift. */
#define APPLY_CARRY_BIT(wp, msize, c, carryshift)               \
  do {                                                          \
    mp_ptr     __wp = (wp);                                     \
    mp_size_t  __msize = (msize);                               \
    unsigned   __carryshift = (carryshift);                     \
    mp_limb_t  __c;                                             \
    ASSERT ((carryshift) >= 0);                                 \
    ASSERT ((carryshift) < GMP_NUMB_BITS);                      \
    ASSERT ((c) == 0 || (c) == 1);                              \
    MPN_ADD_1 (__c, __wp, __wp, __msize, (c) << __carryshift);  \
    MPN_INCR_U (__wp, __msize, __c << __carryshift);            \
  } while (0)

/* {wp,msize} -= c << carryshift, where c is a borrow bit, 0 or 1.
   If there's a borrow out of that subtraction, it's subtracted in turn at
   the low end the same way, ie. newc << carryshift. */
#define APPLY_BORROW_BIT(wp, msize, c, carryshift)              \
  do {                                                          \
    mp_ptr     __wp = (wp);                                     \
    mp_size_t  __msize = (msize);                               \
    unsigned   __carryshift = (carryshift);                     \
    mp_limb_t  __c;                                             \
    ASSERT ((carryshift) >= 0);                                 \
    ASSERT ((carryshift) < GMP_NUMB_BITS);                      \
    ASSERT ((c) == 0 || (c) == 1);                              \
    MPN_SUB_1 (__c, __wp, __wp, __msize, (c) << __carryshift);  \
    MPN_DECR_U (__wp, __msize, __c << __carryshift);            \
  } while (0)


/* {wp,msize} += c << carryshift, where c is any limb NUMB value.
   If there's a carry out of that addition, it's added in at the low end the
   same way, ie. newc << carryshift. */
#define APPLY_CARRY(wp, msize, c, carryshift)           \
  do {                                                  \
    mp_ptr     __wp = (wp);                             \
    mp_size_t  __msize = (msize);                       \
    unsigned   __carryshift = (carryshift);             \
    mp_limb_t  __c, __chi, __clo;                       \
                                                        \
    ASSERT ((carryshift) >= 0);                         \
    ASSERT ((carryshift) < GMP_NUMB_BITS);              \
    ASSERT ((msize) >= 2);                              \
                                                        \
    LSHIFT_21_ANY (__chi, __clo, c, __carryshift);      \
    MPN_ADD_2 (__c, __wp, __wp, __msize, __chi, __clo); \
    MPN_INCR_U (__wp, __msize, __c << __carryshift);    \
  } while (0)


/* Same as APPLY_CARRY, but knowing carryshift>0. */
#define APPLY_CARRY_CARRYSHIFT(wp, msize, c, carryshift)        \
  do {                                                          \
    mp_ptr     __wp = (wp);                                     \
    mp_size_t  __msize = (msize);                               \
    unsigned   __carryshift = (carryshift);                     \
    mp_limb_t  __c, __chi, __clo;                               \
                                                                \
    ASSERT ((carryshift) > 0);                                  \
    ASSERT ((carryshift) < GMP_NUMB_BITS);                      \
    ASSERT ((msize) >= 2);                                      \
                                                                \
    LSHIFT_21 (__chi, __clo, c, __carryshift);                  \
    MPN_ADD_2 (__c, __wp, __wp, __msize, __chi, __clo);         \
    MPN_INCR_U (__wp, __msize, __c << __carryshift);            \
  } while (0)

/* Same as APPLY_BORROW, but knowing carryshift>0. */
#define APPLY_BORROW_CARRYSHIFT(wp, msize, c, carryshift)       \
  do {                                                          \
    mp_ptr     __wp = (wp);                                     \
    mp_size_t  __msize = (msize);                               \
    unsigned   __carryshift = (carryshift);                     \
    mp_limb_t  __c, __chi, __clo;                               \
                                                                \
    ASSERT ((carryshift) > 0);                                  \
    ASSERT ((carryshift) < GMP_NUMB_BITS);                      \
    ASSERT ((msize) >= 2);                                      \
                                                                \
    LSHIFT_21 (__chi, __clo, c, __carryshift);                  \
    MPN_SUB_2 (__c, __wp, __wp, __msize, __chi, __clo);         \
    MPN_DECR_U (__wp, __msize, __c << __carryshift);            \
  } while (0)

/* Same as APPLY_CARRY, with carryshift==0. */
#define APPLY_CARRY_NOCARRYSHIFT(wp, msize, c)  \
  do {                                          \
    mp_ptr     __wp = (wp);                     \
    mp_size_t  __msize = (msize);               \
    mp_limb_t  __c;                             \
    MPN_ADD_1 (__c, __wp, __wp, __msize, c);    \
    MPN_INCR_U (__wp, __msize, __c);            \
  } while (0)

/* Same as APPLY_BORROW, with carryshift==0. */
#define APPLY_BORROW_NOCARRYSHIFT(wp, msize, c) \
  do {                                          \
    mp_ptr     __wp = (wp);                     \
    mp_size_t  __msize = (msize);               \
    mp_limb_t  __c;                             \
    MPN_SUB_1 (__c, __wp, __wp, __msize, c);    \
    MPN_DECR_U (__wp, __msize, __c);            \
  } while (0)


static void
mpr_2expsub1__add (mpr_ptr w, mpr_srcptr x, mpr_srcptr y, mpm_srcptr m)
{
  mp_size_t  msize;
  mp_ptr     wp;
  mp_limb_t  c;

  ASSERT_MPR_2EXPSUB1 (x, m);
  ASSERT_MPR_2EXPSUB1 (y, m);

  msize = MRSIZE(m);
  wp = RPTR (w);
  c = mpn_add_n (wp, RPTR(x), RPTR(y), msize);
  APPLY_CARRY_BIT (wp, msize, c, MCARRYSHIFT(m));

  ASSERT_MPR_2EXPSUB1 (w, m);
}

static void
mpr_2expsub1__add_ui (mpr_ptr w, mpr_srcptr x, unsigned long y, mpm_srcptr m)
{
  mp_size_t  msize;
  mp_ptr     wp;
  mp_limb_t  c;

  ASSERT_MPR_2EXPSUB1 (x, m);

  msize = MRSIZE(m);
  wp = RPTR(w);
#if BITS_PER_ULONG <= GMP_NUMB_BITS
  MPN_ADD_1 (c, wp, RPTR(x), msize, (mp_limb_t) y);
#else
  MPN_ADD_2 (c, wp, RPTR(x), msize,
             (mp_limb_t) y >> GMP_NUMB_BITS, (mp_limb_t) y & GMP_NUMB_MASK);
#endif
  APPLY_CARRY_BIT (wp, msize, c, MCARRYSHIFT(m));

  ASSERT_MPR_2EXPSUB1 (w, m);
}

static void
mpr_2expsub1__addmul_ui__shift (mpr_ptr w, mpr_srcptr x, unsigned long y, mpm_srcptr m)
{
  mp_size_t  msize;
  mp_ptr     wp;
  mp_srcptr  xp;
  mp_limb_t  c;
  unsigned   carryshift;

  ASSERT_MPR_2EXPSUB1 (x, m);
  ASSERT (MRSIZE(m) >= 2);

  xp = RPTR(x);
  wp = RPTR(w);
  msize = MRSIZE(m);
  carryshift = MCARRYSHIFT(m);

#if BITS_PER_ULONG > GMP_NUMB_BITS
  if (UNLIKELY (y > GMP_NUMB_MAX))
    {
      mp_limb_t  yp[2], c2, c1, c0;
      yp[0] = (mp_limb_t) y & GMP_NUMB_MASK;
      yp[1] = (mp_limb_t) y >> GMP_NUMB_BITS;

      wp[msize] = 0;
      MPN_ADDMUL_2 (c, wp, xp, msize, yp);
      LSHIFT_32 (c2,c1,c0, c,wp[msize], carryshift);

      if (msize >= 3)
        {
          MPN_ADD_3 (c, wp, wp, msize, c2, c1, c0);
          APPLY_CARRY_BIT (wp, msize, c, carryshift);
        }
      else
        {
          ADDC_LIMB (c, wp[0], wp[0], c0);
          ADDE_LIMB (c, wp[1], wp[1], c1, c);
          c += c2;
          ASSERT (c >= c2);
          APPLY_CARRY_CARRYSHIFT (wp, msize, c, carryshift);
        }
    }
  else
#endif
    {
      c = mpn_addmul_1 (wp, xp, msize, (mp_limb_t) y);
      APPLY_CARRY_CARRYSHIFT (wp, msize, c, MCARRYSHIFT(m));
    }

  ASSERT_MPR_2EXPSUB1 (w, m);
}

static void
mpr_2expsub1__addmul_ui__noshift (mpr_ptr w, mpr_srcptr x, unsigned long y, mpm_srcptr m)
{
  mp_size_t  msize;
  mp_ptr     wp;
  mp_srcptr  xp;
  mp_limb_t  c;

  ASSERT_MPR_2EXPSUB1 (x, m);
  ASSERT (MRSIZE(m) >= 2);

  wp = RPTR(w);
  xp = RPTR(x);
  msize = MRSIZE(m);

#if BITS_PER_ULONG > GMP_NUMB_BITS
  if (UNLIKELY (y > GMP_NUMB_MAX))
    {
      mp_limb_t  yp[2];
      yp[0] = (mp_limb_t) y & GMP_NUMB_MASK;
      yp[1] = (mp_limb_t) y >> GMP_NUMB_BITS;
      wp[msize] = 0;
      MPN_ADDMUL_2 (c, wp, xp, msize, yp);
      MPN_ADD_2 (c, wp, wp, msize, c,wp[msize]);
      APPLY_CARRY_BIT (wp, msize, c, 0);
    }
  else
#endif
    {
      c = mpn_addmul_1 (wp, xp, msize, (mp_limb_t) y);
      APPLY_CARRY_NOCARRYSHIFT (wp, msize, c);
    }

  ASSERT_MPR_2EXPSUB1 (w, m);
}

static void
mpr_2expsub1__div_2exp (mpr_ptr w, mpr_srcptr x, unsigned long e, mpm_srcptr m)
{
  unsigned long  n;

  n = MBITS(m);
  if (UNLIKELY (e >= n))
    e %= n;

  if (e != 0)
    e = n - e;

  mpr_mul_2exp (w, x, e, m);
}

/* ENHANCE-ME: Do something with calculating a difference abs(x-y) from low
   to high and testing for that a multiple of m, per
   mpr_2expsub1__equal_zero.  */
static int
mpr_2expsub1__equal (mpr_srcptr x, mpr_srcptr y, mpm_srcptr m)
{
  mpz_t  xz;
  mpz_t  yz;
  MPZ_FAKE_R (xz, x, m);
  MPZ_FAKE_R (yz, y, m);
  return mpz_congruent_p (xz, yz, MZ(m));
}

/* ENHANCE-ME: Do something like mpr_2expsub1__equal but knowing y==-1. */
static int
mpr_2expsub1__equal_negone (mpr_srcptr x, mpm_srcptr m)
{
  mpz_t  xz;
  MPZ_FAKE_SI_VAR (yz, -1L);
  MPZ_FAKE_R (xz, x, m);
  return mpz_congruent_p (xz, yz, MZ(m));
}

/* ENHANCE-ME: Do something like mpr_2expsub1__equal but knowing y==1. */
static int
mpr_2expsub1__equal_one (mpr_srcptr x, mpm_srcptr m)
{
  mpz_t  xz;
  MPZ_FAKE_SI_VAR (yz, 1L);
  MPZ_FAKE_R (xz, x, m);
  return mpz_congruent_p (xz, yz, MZ(m));
}

/* ENHANCE-ME: Follow mpr_2expsub1__equal but knowing y small. */
static int
mpr_2expsub1__equal_ui (mpr_srcptr x, unsigned long y, mpm_srcptr m)
{
  mpz_t  xz;
  MPZ_FAKE_UI_VAR (yz, y);
  MPZ_FAKE_R (xz, x, m);
  return mpz_congruent_p (xz, yz, MZ(m));
}

/* ENHANCE-ME: For msize>=2 the low limb of x will say what multiple of m we
   should have, if we're indeed equal to zero, then can expect FFs in the
   middle limbs and a suitable high limb.  */
static int
mpr_2expsub1__equal_zero (mpr_srcptr x, mpm_srcptr m)
{
  mp_size_t  msize, xsize;
  mp_srcptr  xp;

  ASSERT_MPR_2EXPSUB1 (x, m);

  msize = MRSIZE(m);
  xp = RPTR(x);
  xsize = msize;
  MPN_NORMALIZE (xp, xsize);
  return mpn_divisible_p (xp, xsize, MPTR(m), msize);
}

static void
mpr_2expsub1__get_z__shift (mpz_ptr w, mpr_srcptr x, mpm_srcptr m)
{
  mp_size_t  msize, i, wsize;
  mp_ptr     wp;
  mp_srcptr  xp;
  mp_limb_t  c, xhigh, whigh, mask;
  unsigned   carryshift, rshift;

  ASSERT_MPR_2EXPSUB1 (x, m);
  ASSERT (MCARRYSHIFT(m) != 0);

  msize = MRSIZE(m);
  MPZ_REALLOC_PTR (wp, w, msize);

  xp = RPTR(x);
  MPN_COPY (wp, xp, msize-1);

  carryshift = MCARRYSHIFT(m);
  mask = (GMP_NUMB_MAX >> carryshift);
  rshift = GMP_NUMB_BITS - carryshift;

  xhigh = xp[msize-1];
  whigh = xhigh & mask;
  c = xhigh >> rshift;
  MPN_ADD_1 (c, wp, wp, msize-1, c);

  whigh += c;
  c = whigh >> rshift;
  ASSERT (c == 0 || c == 1);
  wp[msize-1] = whigh & mask;
  MPN_INCR_U (wp, msize, c);

  for (i = 0; i < msize-1; i++)
    if (LIKELY (wp[i] != GMP_NUMB_MAX))
      goto not_ffs;
  if (LIKELY (wp[msize-1] != mask))
    goto not_ffs;

  /* w is 2^n-1, which means zero */
  SIZ(w) = 0;
  goto done;

 not_ffs:
  wsize = msize;
  MPN_NORMALIZE (wp, wsize);
  SIZ(w) = wsize;

 done:
  ASSERT_MPR_GET_Z (w, m);
}

/* No excess high bits, so x is either in range already or is 0xFF..FF which
   is zero. */
static void
mpr_2expsub1__get_z__noshift (mpz_ptr w, mpr_srcptr x, mpm_srcptr m)
{
  mp_size_t  msize, i, wsize;
  mp_ptr     wp;

  ASSERT_MPR_2EXPSUB1 (x, m);
  ASSERT (MCARRYSHIFT(m) == 0);

  msize = MRSIZE(m);
  MPZ_REALLOC_PTR (wp, w, msize);

  MPN_COPY (wp, RPTR(x), msize);

  for (i = 0; i < msize; i++)
    if (LIKELY (wp[i] != GMP_NUMB_MAX))
      goto not_ffs;

  /* w is 2^n-1, which means zero */
  SIZ(w) = 0;
  goto done;

 not_ffs:
  wsize = msize;
  MPN_NORMALIZE (wp, wsize);
  SIZ(w) = wsize;

 done:
  ASSERT_MPR_GET_Z (w, m);
}

void
mpr_2expsub1__mul__shift (mpr_ptr w, mpr_srcptr x, mpr_srcptr y, mpm_srcptr m)
{
  mp_size_t  msize;
  mp_ptr     pp, hp, wp;
  int        carryshift;
  mp_limb_t  c;
  TMP_DECL (marker);

  ASSERT_MPR_2EXPSUB1 (x, m);
  ASSERT_MPR_2EXPSUB1 (y, m);
  ASSERT (MCARRYSHIFT(m) != 0);
  ASSERT (MRSIZE(m) >= 2);

  TMP_MARK (marker);
  msize = MRSIZE(m);
  pp = TMP_ALLOC_LIMBS (2*msize);

  mpn_mul_n (pp, RPTR(x), RPTR(y), msize);

  carryshift = MCARRYSHIFT(m);
  hp = pp + msize;
  c = mpn_lshift (hp, hp, msize, carryshift);

  wp = RPTR(w);
  c += mpn_add_n (wp, pp, hp, msize);
  APPLY_CARRY_CARRYSHIFT (wp, msize, c, carryshift);

  ASSERT_MPR_2EXPSUB1 (w, m);
  TMP_FREE (marker);
}

void
mpr_2expsub1__mul__noshift (mpr_ptr w, mpr_srcptr x, mpr_srcptr y, mpm_srcptr m)
{
  mp_size_t  msize;
  mp_ptr     pp, wp;
  mp_limb_t  c;
  TMP_DECL (marker);

  ASSERT_MPR_2EXPSUB1 (x, m);
  ASSERT_MPR_2EXPSUB1 (y, m);
  ASSERT (MCARRYSHIFT(m) == 0);

  TMP_MARK (marker);
  msize = MRSIZE(m);
  pp = TMP_ALLOC_LIMBS (2*msize);

  mpn_mul_n (pp, RPTR(x), RPTR(y), msize);

  wp = RPTR(w);
  c = mpn_add_n (wp, pp, pp + msize, msize);
  APPLY_CARRY_BIT (wp, msize, c, 0);

  ASSERT_MPR_2EXPSUB1 (w, m);
  TMP_FREE (marker);
}


/* mul_2exp is essentially a rotate within n bits, meaning the low e bits go
   to the high end and the high n-e go to the low end.  The carryshift
   excess bits in the high limb go to bit position e in the result, which is
   an overlap so they get added.

   But the implementation treats the data in whole limbs, so the low
   msize-elimbs many limbs are moved up by elimbs and lshifted by ebits.
   ctop is the carry out of this and is applied in the usual way for a carry
   out of the top end of the data.

   The high elimbs are similarly lshifted into the low of the result, but
   they've wrapped around and hence get an extra carryshift,
   ie. lowebits=ebits+carryshift.  cmid is the carry out of this, it
   overlaps with the data already written and so gets added.  (Because
   lowebits is more than ebits it can't be just bitwise-ORed, it must be
   added.

   lowebits can be more than a limb, in which case there's a whole limb of
   overlap, plus the carry out of the shift.  This is treated by fetching
   the overlapping limb out of the destination and letting mpn_lshift write
   that, then adding it back in.  */

static void
mpr_2expsub1__mul_2exp__shift (mpr_ptr w, mpr_srcptr x, unsigned long e, mpm_srcptr m)
{
  unsigned long  n;
  mp_ptr         wp;
  mp_srcptr      xp;
  mp_size_t      msize, elimbs;
  unsigned       ebits, lowebits, carryshift;
  mp_limb_t      ctop;
  TMP_DECL (marker);

  ASSERT_MPR_2EXPSUB1 (x, m);
  ASSERT (MRSIZE(m) >= 2);
  ASSERT (MCARRYSHIFT(m) != 0);

  /* expect small e usually, so avoid "%" which can be slow
     ENHANCE-ME: could calculate an n_preinv, if that helped */
  n = MBITS(m);
  if (UNLIKELY (e >= n))
    e %= n;

  wp = RPTR(w);
  xp = RPTR(x);
  msize = MRSIZE(m);

  TMP_MARK (marker);
  if (w == x)
    {
      mp_ptr  tp;
      tp = TMP_ALLOC_LIMBS (msize);
      MPN_COPY (tp, xp, msize);
      xp = tp;
    }

  elimbs = e / GMP_NUMB_BITS;
  ebits = e % GMP_NUMB_BITS;

  /* low msize-elimbs of x go to high of w */
  if (ebits == 0)
    {
      MPN_COPY (wp+elimbs, xp, msize-elimbs);
      ctop = 0;
    }
  else
    ctop = mpn_lshift (wp+elimbs, xp, msize-elimbs, ebits);

  /* high elimbs of x go to low of w, with extra carryshift */
  carryshift = MCARRYSHIFT(m);
  lowebits = ebits + carryshift;
  if (elimbs != 0)
    {
      mp_limb_t  cmid, ctop2;
      if (lowebits < GMP_NUMB_BITS)
        {
          cmid = mpn_lshift (wp, xp+msize-elimbs, elimbs, lowebits);
        }
      else
        {
          cmid = wp[elimbs];  /* overlapping limb will be added */

          wp[0] = 0;
          lowebits -= GMP_NUMB_BITS;
          if (lowebits == 0)
            MPN_COPY (wp+1, xp+msize-elimbs, elimbs);
          else
            {
              mp_limb_t  cmid2;
              cmid2 = mpn_lshift (wp+1, xp+msize-elimbs, elimbs, lowebits);
              MPN_ADD_1 (ctop2, wp+1+elimbs, wp+1+elimbs, msize-(elimbs+1),
                         cmid2);
              ctop += ctop2;
              ASSERT (ctop >= ctop2);
            }
        }
      MPN_ADD_1 (ctop2, wp+elimbs, wp+elimbs, msize-elimbs, cmid);
      ctop += ctop2;
      ASSERT (ctop >= ctop2);
    }

  APPLY_CARRY_CARRYSHIFT (wp, msize, ctop, carryshift);

  TMP_FREE (marker);
  ASSERT_MPR_2EXPSUB1 (w, m);
}

/* With no carryshift to worry about, ie. n is a multiple of GMP_NUMB_BITS,
   this becomes a much simpler rotation within msize limbs.  */

static void
mpr_2expsub1__mul_2exp__noshift (mpr_ptr w, mpr_srcptr x, unsigned long e, mpm_srcptr m)
{
  unsigned long  n;
  mp_ptr         wp;
  mp_srcptr      xp;
  mp_size_t      msize, elimbs;
  unsigned       ebits;
  mp_limb_t      ctop, cmid;
  TMP_DECL (marker);

  ASSERT_MPR_2EXPSUB1 (x, m);
  ASSERT (MRSIZE(m) >= 2);
  ASSERT (MCARRYSHIFT(m) == 0);

  /* expect small e usually, so avoid "%" which can be slow */
  n = MBITS(m);
  if (UNLIKELY (e >= n))
    e %= n;

  wp = RPTR(w);
  xp = RPTR(x);
  msize = MRSIZE(m);

  TMP_MARK (marker);
  if (w == x)
    {
      mp_ptr  tp;
      tp = TMP_ALLOC_LIMBS (msize);
      MPN_COPY (tp, xp, msize);
      xp = tp;
    }

  elimbs = e / GMP_NUMB_BITS;
  ebits = e % GMP_NUMB_BITS;

  /* low msize-elimbs of x go to high of w */
  if (ebits == 0)
    {
      MPN_COPY (wp+elimbs, xp, msize-elimbs);
      MPN_COPY (wp, xp+msize-elimbs, elimbs);
    }
  else
    {
      ctop = mpn_lshift (wp+elimbs, xp, msize-elimbs, ebits);
      if (elimbs != 0)
        {
          cmid = mpn_lshift (wp, xp+msize-elimbs, elimbs, ebits);
          wp[elimbs] |= cmid;
        }
      wp[0] |= ctop;
    }

  TMP_FREE (marker);
  ASSERT_MPR_2EXPSUB1 (w, m);
}


static void
mpr_2expsub1__mul_si (mpr_ptr w, mpr_srcptr x, long y, mpm_srcptr m)
{
  if (y < 0)
    {
      mpr_2expsub1__neg (w, x, m);
      y = -y;
      x = w;
    }
  mpr_mul_ui (w, x, y, m);
}

static void
mpr_2expsub1__mul_ui__shift (mpr_ptr w, mpr_srcptr x, unsigned long y, mpm_srcptr m)
{
  mp_size_t  msize;
  mp_ptr     wp;
  mp_srcptr  xp;
  mp_limb_t  c;
  unsigned   carryshift;

  ASSERT_MPR_2EXPSUB1 (x, m);
  ASSERT (MRSIZE(m) >= 2);
  ASSERT (MCARRYSHIFT(m) != 0);

  wp = RPTR(w);
  xp = RPTR(x);
  msize = MRSIZE(m);
  carryshift = MCARRYSHIFT(m);

#if BITS_PER_ULONG > GMP_NUMB_BITS
  if (UNLIKELY (y > GMP_NUMB_MAX))
    {
      mp_limb_t  yp[2], c2, c1, c0;
      yp[0] = (mp_limb_t) y & GMP_NUMB_MASK;
      yp[1] = (mp_limb_t) y >> GMP_NUMB_BITS;

      MPN_MUL_2 (c, wp, xp, msize, yp);
      LSHIFT_32 (c2,c1,c0, c,wp[msize], carryshift);

      if (msize >= 3)
        {
          MPN_ADD_3 (c, wp, wp, msize, c2, c1, c0);
          APPLY_CARRY_BIT (wp, msize, c, carryshift);
        }
      else
        {
          ADDC_LIMB (c, wp[0], wp[0], c0);
          ADDE_LIMB (c, wp[1], wp[1], c1, c);
          c += c2;
          ASSERT (c >= c2);
          APPLY_CARRY_CARRYSHIFT (wp, msize, c, carryshift);
        }
    }
  else
#endif
    {
      c = mpn_mul_1 (wp, xp, msize, (mp_limb_t) y);
      APPLY_CARRY_CARRYSHIFT (wp, msize, c, carryshift);
    }

  ASSERT_MPR_2EXPSUB1 (w, m);
}

static void
mpr_2expsub1__mul_ui__noshift (mpr_ptr w, mpr_srcptr x, unsigned long y, mpm_srcptr m)
{
  mp_size_t  msize;
  mp_ptr     wp;
  mp_srcptr  xp;
  mp_limb_t  c;

  ASSERT_MPR_2EXPSUB1 (x, m);
  ASSERT (MRSIZE(m) >= 2);
  ASSERT (MCARRYSHIFT(m) == 0);

  wp = RPTR(w);
  xp = RPTR(x);
  msize = MRSIZE(m);

#if BITS_PER_ULONG > GMP_NUMB_BITS
  if (UNLIKELY (y > GMP_NUMB_MAX))
    {
      mp_limb_t  yp[2];
      yp[0] = (mp_limb_t) y & GMP_NUMB_MASK;
      yp[1] = (mp_limb_t) y >> GMP_NUMB_BITS;
      MPN_MUL_2 (c, wp, xp, msize, yp);
      MPN_ADD_2 (c, wp, wp, msize, c,wp[msize]);
      APPLY_CARRY_BIT (wp, msize, c, 0);
    }
  else
#endif
    {
      c = mpn_mul_1 (wp, xp, msize, (mp_limb_t) y);
      APPLY_CARRY_NOCARRYSHIFT (wp, msize, c);
    }

  ASSERT_MPR_2EXPSUB1 (w, m);
}

static void
mpr_2expsub1__neg (mpr_ptr w, mpr_srcptr x, mpm_srcptr m)
{
  mp_size_t  msize;
  mp_ptr     wp;
  mp_limb_t  c;

  ASSERT_MPR_2EXPSUB1 (x, m);

  msize = MRSIZE(m);
  wp = RPTR(w);
  mpn_com_n (wp, RPTR(x), msize);
  MPN_ADD_1 (c, wp, wp, msize, CNST_LIMB(1));
  APPLY_BORROW_BIT (wp, msize, CNST_LIMB(1)-c, MCARRYSHIFT(m));

  ASSERT_MPR_2EXPSUB1 (w, m);
}

/* ENHANCE-ME: Sliding window, perhaps with on-demand calculation of
   intermediate powers.  */
static void
mpr_2expsub1__pow__shift (mpr_ptr r, mpr_srcptr b, mpz_srcptr e, mpm_srcptr m)
{
  mp_size_t  msize, esize, i;
  mp_ptr     rp, pp, hp;
  mp_srcptr  bp, ep, minv, mp;
  mp_limb_t  elimb, mask, c;
  int        lz;
  unsigned   carryshift;
  TMP_DECL (marker);

  TRACE (printf ("mpr_2expsub1__pow__noshift"));

  ASSERT_MPR_2EXPSUB1 (b, m);
  ASSERT_MPZ (e);

  TMP_MARK (marker);
  msize = MRSIZE(m);
  rp = RPTR(r);

  esize = SIZ(e);
  if (esize == 0)
    {
      /* b^0 == 1 */
      rp[0] = 1;
      MPN_ZERO (rp+1, msize-1);
      return;
    }

  bp = RPTR(b);
  if (esize < 0)
    {
      esize = -esize;
      if (! mpr_gen__invert (r, b, m))
        DIVIDE_BY_ZERO;
      bp = rp;
    }

  TMP_ALLOC_LIMBS_2SEP (pp, 2*msize, rp, bp, msize);

  MPN_COPY (rp, bp, msize);  /* start b^1 */
  mp = MPTR(m);
  minv = MINV(m);
  carryshift = MCARRYSHIFT(m);
  hp = pp + msize;

  ep = PTR(e);
  i = esize-1;
  elimb = ep[i];
  count_leading_zeros (lz, elimb);
  mask = CNST_LIMB(1) << (BITS_PER_MP_LIMB - 1 - lz);

  for (;;)
    {
      mask >>= 1;
      if (mask == 0)
        {
          i--;
          if (i < 0)
            break;
          elimb = ep[i];
          mask = GMP_NUMB_HIGHBIT;
        }
      TRACE (printf (" at i=%ld mask=%lX bit=%d\n",
                     i, mask, (elimb & mask) != 0));

      mpn_sqr_n (pp, rp, msize);
      c = mpn_lshift (hp, hp, msize, carryshift);
      c += mpn_add_n (rp, pp, hp, msize);
      APPLY_CARRY_CARRYSHIFT (rp, msize, c, carryshift);

      if ((elimb & mask) != 0)
        {
          mpn_mul_n (pp, rp, bp, msize);
          c = mpn_lshift (hp, hp, msize, carryshift);
          c += mpn_add_n (rp, pp, hp, msize);
          APPLY_CARRY_CARRYSHIFT (rp, msize, c, carryshift);
        }
    }

  ASSERT_MPR_2EXPSUB1 (r, m);
  TMP_FREE (marker);
}

/* ENHANCE-ME: Sliding window, perhaps with on-demand calculation of
   intermediate powers.  */
static void
mpr_2expsub1__pow__noshift (mpr_ptr r, mpr_srcptr b, mpz_srcptr e, mpm_srcptr m)
{
  mp_size_t  msize, esize, i;
  mp_ptr     rp, pp;
  mp_srcptr  bp, ep, minv, mp;
  mp_limb_t  elimb, mask, c;
  int        lz;
  TMP_DECL (marker);

  ASSERT_MPR_2EXPSUB1 (b, m);
  ASSERT_MPZ (e);
  ASSERT (MCARRYSHIFT(m) == 0);

  TMP_MARK (marker);
  msize = MRSIZE(m);
  rp = RPTR(r);

  esize = SIZ(e);
  if (esize == 0)
    {
      /* b^0 == 1 */
      rp[0] = 1;
      MPN_ZERO (rp+1, msize-1);
      return;
    }

  bp = RPTR(b);
  if (esize < 0)
    {
      esize = -esize;
      if (! mpr_gen__invert (r, b, m))
        DIVIDE_BY_ZERO;
      bp = rp;
    }

  TMP_ALLOC_LIMBS_2SEP (pp, 2*msize, rp, bp, msize);

  MPN_COPY (rp, bp, msize);  /* start b^1 */
  mp = MPTR(m);
  minv = MINV(m);

  ep = PTR(e);
  i = esize-1;
  elimb = ep[i];
  count_leading_zeros (lz, elimb);
  mask = CNST_LIMB(1) << (BITS_PER_MP_LIMB - 1 - lz);

  for (;;)
    {
      mask >>= 1;
      if (mask == 0)
        {
          i--;
          if (i < 0)
            break;
          elimb = ep[i];
          mask = GMP_NUMB_HIGHBIT;
        }
      TRACE (printf (" at i=%ld mask=%lX bit=%d\n",
                     i, mask, (elimb & mask) != 0));

      mpn_sqr_n (pp, rp, msize);
      c = mpn_add_n (rp, pp, pp+msize, msize);
      APPLY_CARRY_BIT (rp, msize, c, 0);

      if ((elimb & mask) != 0)
        {
          mpn_mul_n (pp, rp, bp, msize);
          c = mpn_add_n (rp, pp, pp+msize, msize);
          APPLY_CARRY_BIT (rp, msize, c, 0);
        }
    }

  ASSERT_MPR_2EXPSUB1 (r, m);
  TMP_FREE (marker);
}

/* ENHANCE-ME: Specific code here for the negative case. */
static void
mpr_2expsub1__set_si (mpr_ptr w, long x, mpm_srcptr m)
{
  if (x >= 0)
    mpr_gen__set_ui__fits (w, x, m);
  else
    {
      mpr_gen__set_ui__fits (w, -x, m);
      mpr_2expsub1__neg (w, w, m);
    }
}

static void
mpr_2expsub1__sub (mpr_ptr w, mpr_srcptr x, mpr_srcptr y, mpm_srcptr m)
{
  mp_size_t  msize;
  mp_limb_t  c;
  mp_ptr     wp;

  ASSERT_MPR_2EXPSUB1 (x, m);
  ASSERT_MPR_2EXPSUB1 (y, m);

  msize = MRSIZE(m);
  wp = RPTR (w);
  c = mpn_sub_n (wp, RPTR(x), RPTR(y), msize);
  APPLY_BORROW_BIT(wp, msize, c, MCARRYSHIFT(m));

  ASSERT_MPR_2EXPSUB1 (w, m);
}

static void
mpr_2expsub1__sub_ui (mpr_ptr w, mpr_srcptr x, unsigned long y, mpm_srcptr m)
{
  mp_size_t  msize;
  mp_ptr     wp;
  mp_limb_t  c;

  ASSERT_MPR_2EXPSUB1 (x, m);

  msize = MRSIZE(m);
  wp = RPTR(w);
#if BITS_PER_ULONG > GMP_NUMB_BITS
  MPN_SUB_2 (c, wp, RPTR(x), msize,
             (mp_limb_t) y >> GMP_NUMB_BITS, (mp_limb_t) y & GMP_NUMB_MASK);
#else
  MPN_SUB_1 (c, wp, RPTR(x), msize, (mp_limb_t) y);
#endif
  APPLY_BORROW_BIT (wp, msize, c, MCARRYSHIFT(m));

  ASSERT_MPR_2EXPSUB1 (w, m);
}

static void
mpr_2expsub1__submul_ui__shift (mpr_ptr w, mpr_srcptr x, unsigned long y, mpm_srcptr m)
{
  mp_size_t  msize;
  mp_ptr     wp;
  mp_srcptr  xp;
  mp_limb_t  c;
  unsigned   carryshift;

  ASSERT_MPR_2EXPSUB1 (x, m);
  ASSERT (MRSIZE(m) >= 2);

  xp = RPTR(x);
  wp = RPTR(w);
  msize = MRSIZE(m);
  carryshift = MCARRYSHIFT(m);

#if BITS_PER_ULONG > GMP_NUMB_BITS
  if (UNLIKELY (y > GMP_NUMB_MAX))
    {
      mp_limb_t  yp[2], c2, c1;
      yp[0] = (mp_limb_t) y & GMP_NUMB_MASK;
      yp[1] = (mp_limb_t) y >> GMP_NUMB_BITS;

      wp[msize] = 0;
      MPN_SUBMUL_2 (c, wp, xp, msize, yp);
      APPLY_CARRY_CARRYSHIFT (wp, msize, wp[msize], carryshift);

      LSHIFT_21 (c2,c1, c, carryshift);
      if (msize >= 3)
        {
          MPN_SUB_2 (c, wp+1, wp+1, msize-1, c2, c1);
          APPLY_BORROW_BIT (wp, msize, c, carryshift);
        }
      else
        {
          SUBC_LIMB (c, wp[1], wp[1], c1);
          c += c2;
          ASSERT (c >= c2);
          APPLY_BORROW_CARRYSHIFT (wp, msize, c, carryshift);
        }
    }
  else
#endif
    {
      c = mpn_submul_1 (wp, xp, msize, (mp_limb_t) y);
      APPLY_BORROW_CARRYSHIFT (wp, msize, c, MCARRYSHIFT(m));
    }

  ASSERT_MPR_2EXPSUB1 (w, m);
}

static void
mpr_2expsub1__submul_ui__noshift (mpr_ptr w, mpr_srcptr x, unsigned long y, mpm_srcptr m)
{
  mp_size_t  msize;
  mp_ptr     wp;
  mp_srcptr  xp;
  mp_limb_t  c;

  ASSERT_MPR_2EXPSUB1 (x, m);
  ASSERT (MRSIZE(m) >= 2);

  wp = RPTR(w);
  xp = RPTR(x);
  msize = MRSIZE(m);

#if BITS_PER_ULONG > GMP_NUMB_BITS
  if (UNLIKELY (y > GMP_NUMB_MAX))
    {
      mp_limb_t  yp[2];
      yp[0] = (mp_limb_t) y & GMP_NUMB_MASK;
      yp[1] = (mp_limb_t) y >> GMP_NUMB_BITS;
      wp[msize] = 0;
      MPN_SUBMUL_2 (c, wp, xp, msize, yp);
      APPLY_CARRY_NOCARRYSHIFT (wp, msize, wp[msize]);
      MPN_SUB_1 (c, wp+1, wp+1, msize-1, c);
      APPLY_BORROW_BIT (wp, msize, c, 0);
    }
  else
#endif
    {
      c = mpn_submul_1 (wp, xp, msize, (mp_limb_t) y);
      APPLY_BORROW_NOCARRYSHIFT (wp, msize, c);
    }

  ASSERT_MPR_2EXPSUB1 (w, m);
}


void
mpm_init_2expsub1 (mpm_ptr m, unsigned long n)
{
  mp_size_t      msize;
  mp_ptr         mp;
  unsigned long  highbits;
  int            carryshift;
  mp_limb_t      mask;

  TRACE (printf ("mpm_init_2expsub1 2^%lu-1\n", n));

  /* reject 2^0-1==0 and 2^1-1==1 */
  if (n < 2)
    DIVIDE_BY_ZERO;

  if (UNLIKELY (n <= GMP_NUMB_BITS))
    {
      mpm_init_one (m, MP_LIMB_T_MAX >> (GMP_LIMB_BITS - n));
      return;
    }

  ASSERT_CODE (memset (m, '\0', sizeof (mpm_t)));

  MBITS(m) = n;

  /* n bits, as limbs */
  msize = (n + GMP_NUMB_BITS - 1) / GMP_NUMB_BITS;
  MRSIZE(m) = msize;
  MRALLOC(m) = msize + (BITS_PER_ULONG > GMP_NUMB_BITS);

  carryshift = ((msize * GMP_NUMB_BITS) - n) % n;
  MCARRYSHIFT(m) = carryshift;
  TRACE (printf ("  MCARRYSHIFT=%d\n", MCARRYSHIFT(m)));

  /* bits in the high limb */
  highbits = n - (msize-1) * GMP_NUMB_BITS;
  ASSERT (highbits >= 1);
  ASSERT (highbits <= GMP_NUMB_BITS);

  /* mask for bits in high limb */
  mask = GMP_NUMB_MASK >> (GMP_NUMB_BITS - highbits);

  mp = __GMP_ALLOCATE_FUNC_LIMBS (msize);
  ALLOC(MZ(m)) = msize;
  SIZ(MZ(m)) = msize;
  MPTR(m) = mp;
  MPN_FILL (mp, msize-1, GMP_NUMB_MASK);
  mp[msize-1] = mask;
  TRACE (gmp_printf ("  m=%#ZX\n", MZ(m)));

  m->_mpm_clear = NULL;

  m->_mpr_add          = mpr_2expsub1__add;
  m->_mpr_add_ui       = mpr_2expsub1__add_ui;
  m->_mpr_addmul_ui    = (carryshift != 0 ? mpr_2expsub1__addmul_ui__shift
                          :                 mpr_2expsub1__addmul_ui__noshift);
  m->_mpr_clear        = mpr_gen__clear;
  m->_mpr_div          = mpr_gen__div__invert_mul;
  m->_mpr_div_2exp     = mpr_2expsub1__div_2exp;
  m->_mpr_equal        = mpr_2expsub1__equal;
  m->_mpr_equal_ui     = mpr_2expsub1__equal_ui;
  m->_mpr_equal_one    = mpr_2expsub1__equal_one;
  m->_mpr_equal_negone = mpr_2expsub1__equal_negone;
  m->_mpr_equal_zero   = mpr_2expsub1__equal_zero;
  m->_mpr_gcd          = mpr_gen__gcd__odd;
  m->_mpr_get_z        = (carryshift != 0 ? mpr_2expsub1__get_z__shift
                          :                 mpr_2expsub1__get_z__noshift);

  /* ENHANCE-ME: This is a bit slack, it should be possible to determine the
     low ulong with less work, most of the time.  */
  m->_mpr_get_ui       = mpr_gen__get_ui__get_z;

  m->_mpr_invert       = mpr_gen__invert;
  m->_mpr_jacobi       = mpr_gen__jacobi;
  m->_mpr_mul          = (carryshift != 0 ? mpr_2expsub1__mul__shift
                          :                 mpr_2expsub1__mul__noshift);
  m->_mpr_mul_2exp     = (carryshift != 0 ? mpr_2expsub1__mul_2exp__shift
                          :                 mpr_2expsub1__mul_2exp__noshift);
  m->_mpr_mul_ui       = (carryshift != 0 ? mpr_2expsub1__mul_ui__shift
                          :                 mpr_2expsub1__mul_ui__noshift);
  m->_mpr_mul_si       = mpr_2expsub1__mul_si;
  m->_mpr_neg          = mpr_2expsub1__neg;
  m->_mpr_new          = mpr_gen__new;
  m->_mpr_pow          = (carryshift != 0 ? mpr_2expsub1__pow__shift
                          :                 mpr_2expsub1__pow__noshift);
  m->_mpr_set          = mpr_gen__set;
  m->_mpr_set_si       = mpr_2expsub1__set_si;
  m->_mpr_set_ui       = mpr_gen__set_ui__fits;
  m->_mpr_set_z        = mpr_mod__set_z;   /* ENHANCE-ME: own version ? */
  m->_mpr_sub          = mpr_2expsub1__sub;
  m->_mpr_sub_ui       = mpr_2expsub1__sub_ui;
  m->_mpr_submul_ui    = (carryshift != 0 ? mpr_2expsub1__submul_ui__shift
                          :                 mpr_2expsub1__submul_ui__noshift);
  m->_mpr_swap         = mpr_gen__swap;
  m->_mpr_urandomm     = mpr_mod__urandomm;
}
