/* Shared functions for speed measuring.

   Copyright 2002, 2007 Kevin Ryde

   This file is part of MPM.

   MPM is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2, or (at your option) any later version.

   MPM is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
   FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
   details.

   You should have received a copy of the GNU General Public License along
   with MPM; see the file COPYING.  If not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301
   USA.  */

#include <stdio.h>
#include <math.h>
#include "gmp.h"
#include "gmp-impl.h"
#include "mpm.h"
#include "mpm-impl.h"
#include "speed.h"


mp_size_t  mullowhalf_karatsuba_threshold;
mp_size_t  redc_big_threshold;
mp_size_t  redc_setz_threshold;


void
mpr_set_n (mpr_ptr w, mp_srcptr xp, mp_size_t xsize, mpm_srcptr m)
{
  mpz_t   xz;
  PTR(xz) = (mp_ptr) xp;
  SIZ(xz) = xsize;
  ASSERT_CODE (ALLOC(xz) = ABS(xsize));
  mpr_set_z (w, xz, m);
}

void
mpz_xor_n (mpz_ptr w, mpz_srcptr x, mp_srcptr yp, mp_size_t ysize)
{
  mpz_t   yz;
  PTR(yz) = (mp_ptr) yp;
  SIZ(yz) = ysize;
  ASSERT_CODE (ALLOC(yz) = ABS(ysize));
  mpz_xor (w, x, yz);
}


#define SPEED_MPM_INIT(m,s)  SPEED_RESTRICT_COND (speed_mpm_init (m, s))

int
speed_mpm_init (mpm_ptr m, struct speed_params *s)
{
  mpz_t  mz;

  mpz_init_set_n (mz, s->yp, s->size);
  mpz_xor_n (mz, mz, s->xp, s->size);
  mpz_setbit (mz, s->size * GMP_NUMB_BITS - 1);

  if (s->r == 0)
    {
      mpz_clrbit (mz, 0L);  /* even */
      mpm_init_div (m, mz);
    }
  else if (s->r == 1)
    {
      redc_big_threshold = MP_SIZE_T_MAX;
      mpz_setbit (mz, 0L);  /* odd */
      mpm_init_redc (m, mz);
    }
  else if (s->r == 2)
    {
      if (s->size < 4)
        return 0;
      redc_big_threshold = 1;
      mpz_setbit (mz, 0L);  /* odd */
      mpm_init_redc (m, mz);
    }
  else if (s->r == 3)
    {
      mpm_init_2exp (m, (unsigned long) (GMP_NUMB_BITS * s->size));
    }
  else if (s->r == 4)
    {
      mpm_init_2expsub1 (m, (unsigned long) (GMP_NUMB_BITS * s->size));
    }
  else if (s->r == 99)
    {
      mpz_setbit (mz, 0L);  /* odd */
      mpm_init_redc (m, mz);
    }
  else
    {
      printf ("Invalid algorithm selection\n");
      abort ();
    }

  mpz_clear (mz);
  return 1;
}

void
tweak_noop (mpr_ptr x, mpr_ptr y, mpm_srcptr m)
{
}

void
tweak_invertible (mpr_ptr x, mpr_ptr y, mpm_srcptr m)
{
  mpz_t  g;
  int    i;

  mpz_init (g);
  for (i = 0; i < 1000; i++)
    {
      mpr_gcd (g, y, m);
      if (mpz_cmp_ui (g, 1L) == 0)
        break;
      mpr_add_ui (y, y, 1L, m);
    }
  mpz_clear (g);
}


#define SPEED_ROUTINE_MPR_UNARY(function)       \
  do {                                          \
    mpm_t     m;                                \
    mpr_t     w, x;                             \
    unsigned  i;                                \
    double    t;                                \
                                                \
    SPEED_RESTRICT_COND (s->size >= 1);         \
    SPEED_MPM_INIT (m, s);                      \
    mpr_init (w, m);                            \
    mpr_init (x, m);                            \
    mpr_set_n (x, s->xp, s->size, m);           \
                                                \
    speed_operand_src (s, RPTR(x), s->size);    \
    speed_operand_dst (s, RPTR(w), s->size);    \
    speed_cache_fill (s);                       \
                                                \
    speed_starttime ();                         \
    i = s->reps;                                \
    do                                          \
      function (w, x, m);                       \
    while (--i != 0);                           \
    t = speed_endtime ();                       \
                                                \
    mpr_clear (w, m);                           \
    mpr_clear (x, m);                           \
    mpm_clear (m);                              \
                                                \
    return t;                                   \
  } while (0)

#define SPEED_ROUTINE_MPR_BINARY(function)      \
  SPEED_ROUTINE_MPR_BINARY_TWEAK(function, tweak_noop)

#define SPEED_ROUTINE_MPR_BINARY_TWEAK(function, tweak)       \
  do {                                          \
    mpm_t     m;                                \
    mpr_t     w, x, y;                          \
    unsigned  i;                                \
    double    t;                                \
                                                \
    SPEED_RESTRICT_COND (s->size >= 1);         \
    SPEED_MPM_INIT (m, s);                      \
    mpr_init (w, m);                            \
    mpr_init (x, m);                            \
    mpr_init (y, m);                            \
    mpr_set_n (x, s->xp, s->size, m);           \
    mpr_set_n (y, s->yp, s->size, m);           \
                                                \
    if (tweak != NULL)                          \
      (*tweak) (x, y, m);                       \
                                                \
    speed_operand_src (s, RPTR(x), s->size);    \
    speed_operand_src (s, RPTR(y), s->size);    \
    speed_operand_dst (s, RPTR(w), s->size);    \
    speed_cache_fill (s);                       \
                                                \
    speed_starttime ();                         \
    i = s->reps;                                \
    do                                          \
      function (w, x, y, m);                    \
    while (--i != 0);                           \
    t = speed_endtime ();                       \
                                                \
    mpr_clear (w, m);                           \
    mpr_clear (x, m);                           \
    mpr_clear (y, m);                           \
    mpm_clear (m);                              \
                                                \
    return t;                                   \
  } while (0)


double
speed_mpr_add (struct speed_params *s)
{
  SPEED_ROUTINE_MPR_BINARY (mpr_add);
}

double
speed_mpr_div (struct speed_params *s)
{
  SPEED_ROUTINE_MPR_BINARY_TWEAK (mpr_div, tweak_invertible);
}

double
speed_mpr_invert (struct speed_params *s)
{
  mpm_t     m;
  mpr_t     w, x;
  mpz_t     xz, invz;
  unsigned  i;
  double    t;

  SPEED_RESTRICT_COND (s->size >= 1);
  SPEED_MPM_INIT (m, s);
  mpr_init (w, m);
  mpr_init (x, m);

  /* inverse exists */
  mpz_init_set_n (xz, s->xp, s->size);
  mpz_init (invz);
  while (mpz_invert (invz, xz, mpm_modref(m)) == 0)
    mpz_add_ui (xz, xz, 1L);
  mpz_clear (invz);

  mpr_set_z (x, xz, m);

  speed_operand_src (s, RPTR(x), s->size);
  speed_operand_dst (s, RPTR(w), s->size);
  speed_cache_fill (s);

  speed_starttime ();
  i = s->reps;
  do
    mpr_invert (w, x, m);
  while (--i != 0);
  t = speed_endtime ();

  mpr_clear (w, m);
  mpr_clear (x, m);
  mpm_clear (m);
  mpz_clear (xz);

  return t;
}

double
speed_mpr_mul (struct speed_params *s)
{
  SPEED_ROUTINE_MPR_BINARY (mpr_mul);
}

double
speed_mpr_sub (struct speed_params *s)
{
  SPEED_ROUTINE_MPR_BINARY (mpr_sub);
}


double
speed_mpn_mullowhalf_basecase (struct speed_params *s)
{
  SPEED_ROUTINE_MPN_MUL_N (mpn_mullowhalf_basecase);
}

double
speed_mpn_mullowhalf_kara (struct speed_params *s)
{
  SPEED_ROUTINE_MPN_MUL_N (mpn_mullowhalf_kara);
}

double
speed_mpn_mullowhalf_n (struct speed_params *s)
{
  SPEED_ROUTINE_MPN_MUL_N (mpn_mullowhalf_n);
}
