/* Demo Pollard-Rho factorization using MPM.

   Copyright 2002, 2007 Kevin Ryde

   This file is part of MPM.

   MPM is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2, or (at your option) any later version.

   MPM is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
   FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
   details.

   You should have received a copy of the GNU General Public License along
   with MPM; see the file COPYING.  If not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301
   USA.  */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "gmp.h"

#include "mpm.h"


/* how many differences to accumulate before taking a gcd */
#define DACC  10


void
pollard (mpz_t az, mpz_t N)
{
  mpz_t  dz, g;
  mpm_t  m;
  mpr_t  x, x2, a, d, dacc, one;
  int    dacc_count;
  unsigned long  k, l;

  if (mpz_probab_prime_p (N, 10))
    printf ("this is probably prime, attempting to factorize anyway ...\n");

  mpz_init (dz);
  mpz_init (g);

  mpm_init (m, N);
  mpr_init_set_ui (x, 5L, m);
  mpr_init_set_ui (x2, 2L, m);
  mpr_init_set_z (a, az, m);
  mpr_init (d, m);
  mpr_init_set_ui (one, 1L, m);
  mpr_init_set (dacc, one, m);
  k = 1;
  l = 1;
  dacc_count = 0;

  for (;;)
    {
      mpr_sub (d, x2, x, m);
      mpr_mul (dacc, dacc, d, m);
      dacc_count++;
      if (dacc_count >= DACC)
        {
          mpr_gcd (g, dacc, m);
          if (mpz_cmp_ui (g, 1) != 0)
            break;
          dacc_count = 0;
          mpr_set (dacc, one, m);
        }

      k--;
      if (k == 0)
        {
          l *= 2;
          k = l;
          mpr_set (x2, x, m);
        }
      mpr_mul (x, x, x, m);
      mpr_add (x, x, a, m);
    }

  mpz_divexact (dz, N, g);
  gmp_printf ("%Zd = %Zd (%s prime) * %Zd (%s prime)\n",
              N,
              g, mpz_probab_prime_p (g, 10) ? "probably" : "not",
              dz, mpz_probab_prime_p (dz, 10) ? "probably" : "not");

  mpr_clear (x, m);
  mpr_clear (x2, m);
  mpr_clear (a, m);
  mpr_clear (d, m);
  mpr_clear (dacc, m);
  mpr_clear (one, m);
  mpm_clear (m);

  mpz_clear (dz);
  mpz_clear (g);
}


int
main (int argc, char *argv[])
{
  mpz_t  az, N;
  int    i;

  if (argc < 2)
    {
    usage:
      printf ("Usage: %s [-a <num>] <num>...\n", argv[0]);
      printf ("    -a <num>   Set the \"a\" parameter (default 1)\n");
      printf ("Factorize the numbers on the command line using Pollard-rho.\n");
      printf ("Only one split is found, and the primality of the cofactors indicated.\n");
      return 0;
    }

  mpz_init_set_ui (az, 1L);
  mpz_init (N);

  for (i = 1; i < argc; i++)
    {
      if (strcmp (argv[i], "-a") == 0)
        {
          i++;
          if (i >= argc)
            {
              printf ("Missing -a parameter\n");
              exit (1);
            }
          if (mpz_set_str (az, argv[i], 0) != 0)
            {
              printf ("Invalid -a argument: %s\n", argv[i]);
              exit (1);
            }
          continue;
        }

      if (strcmp (argv[i], "--help") == 0)
        goto usage;

      if (mpz_set_str (N, argv[i], 0) != 0)
        {
          printf ("Invalid number to factorize: %s\n", argv[i]);
          continue;
        }

      pollard (az, N);
    }

  mpz_clear (az);
  mpz_clear (N);

  return 0;
}
