/* mpm division method.

   Copyright 2002, 2007 Kevin Ryde

   This file is part of MPM.

   MPM is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2, or (at your option) any later version.

   MPM is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
   FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
   details.

   You should have received a copy of the GNU General Public License along
   with MPM; see the file COPYING.  If not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301
   USA.  */

#include <stdio.h>
#include "gmp.h"
#include "gmp-impl.h"

#include "mpm.h"
#include "mpm-impl.h"


/* A residue is in the range 0 to m-1.

   An alternative residue form would be to simply require ABSIZ(x) <=
   SIZ(m->_mp_m).  But it's not clear letting residues grow a little like
   that would help.  */


/* change this to "#define TRACE(x) x" for diagnostics */
#define TRACE(x)

#define ASSERT_MPR_DIV(w, m)   ASSERT_MPR_MOD (w, m)



void
mpr_div__add_ui (mpr_ptr w, mpr_srcptr x, unsigned long y, mpm_srcptr m)
{
  mp_size_t  msize;
  mp_ptr     wp;
  mp_srcptr  xp, mp;
  mp_limb_t  c, c2;

  ASSERT_MPR_DIV (x, m);
  ASSERT (MRSIZE(m) >= 2);

  wp = RPTR(w);
  xp = RPTR(x);
  mp = MPTR(m);
  msize = MRSIZE(m);

#if BITS_PER_ULONG > GMP_NUMB_BITS
  if (UNLIKELY (y > GMP_NUMB_MAX))
    {
      mp_limb_t  yhi = y >> GMP_NUMB_BITS;
      mp_limb_t  ylo = y & GMP_NUMB_MASK;
      if (UNLIKELY (msize == 2 && CMP_2_GE (yhi,ylo, mp[1],mp[0])))
        {
          mp_limb_t  ylimb = y;
          ylimb %= (mp[0] + (mp[1] << GMP_NUMB_BITS));
          yhi = ylimb >> GMP_NUMB_BITS;
          ylo = ylimb & GMP_NUMB_MASK;
        }
      MPN_ADD_2 (c, wp, xp, msize, yhi, ylo);
    }
  else
#endif
    MPN_ADD_1 (c, wp, xp, msize, (mp_limb_t) y);

  if (LIKELY (c == 0))
    MPN_CMP_LT_ACTION (wp, mp, msize, goto done);

  c2 = mpn_sub_n (wp, wp, mp, msize);
  ASSERT (c2 == c);

 done:
  ASSERT_MPR_DIV (w, m);
}

int
mpr_div__equal_one (mpr_srcptr x, mpm_srcptr m)
{
  mp_srcptr  xp;

  ASSERT_MPR_DIV (x, m);

  xp = RPTR(x);
  if (xp[0] != 1)
    return 0;
  MPN_NONZERO_P_ACTION (xp+1, MRSIZE(m)-1, return 0);
  return 1;
}

int
mpr_div__equal_ui (mpr_srcptr x, unsigned long y, mpm_srcptr m)
{
  mp_srcptr  xp;
  mp_size_t  i, msize;

  ASSERT_MPR_DIV (x, m);
  ASSERT (MRSIZE(m) >= 2);

  if (UNLIKELY (y > GMP_NUMB_MAX))
    {
      mpr_t  yr;
      int    ret;
      mpr_init_set_ui (yr, y, m);
      ret = mpr_equal (x, yr, m);
      mpr_clear (yr, m);
      return ret;
    }

  xp = RPTR(x);
  if (xp[0] != (y & GMP_NUMB_MASK))
    return 0;
  i = 1;

#if GMP_NUMB_BITS < BITS_PER_ULONG
  if (xp[1] != (y >> GMP_NUMB_BITS))
    return 0;
  i = 2;
#endif

  msize = MRSIZE(m);
  for ( ; i < msize; i++)
    if (xp[i] != 0)
      return 0;
  return 1;
}

/* Calling mpn_gcd directly from here would be possible, but it'd be pretty
   much an exact duplication of mpz_gcd.  */
void
mpr_div__gcd (mpz_ptr w, mpr_srcptr x, mpm_srcptr m)
{
  mpz_t      xz;
  mp_srcptr  xp;
  mp_size_t  xsize;

  ASSERT_MPR_MOD (x, m);

  xp = RPTR(x);
  xsize = MRSIZE(m);
  MPN_NORMALIZE (xp, xsize);

  PTR(xz) = (mp_ptr) xp;
  SIZ(xz) = xsize;
  ASSERT_CODE (ALLOC(xz) = xsize);

  mpz_gcd (w, xz, MZ(m));
}

unsigned long
mpr_div__get_ui (mpr_srcptr x, mpm_srcptr m)
{
  unsigned long  w;
  mp_srcptr      xp;
  mp_size_t      msize;

  ASSERT_MPR_DIV (x, m);
  xp = RPTR(x);
  msize = MRSIZE(m);

  w = xp[0];
#if GMP_NUMB_BITS < BITS_PER_ULONG
  if (msize > 1)
    w |= (xp[1] << GMP_NUMB_BITS);
#endif

  ASSERT_MPR_GET_UI (w, m);
  return w;
}

void
mpr_div__get_z (mpz_ptr w, mpr_srcptr x, mpm_srcptr m)
{
  mp_srcptr  xp;
  mp_size_t  xsize;
  mp_ptr     wp;

  ASSERT_MPR_DIV (x, m);

  xp = RPTR(x);
  xsize = MRSIZE(m);

  MPN_NORMALIZE (xp, xsize);
  SIZ(w) = xsize;
  MPZ_REALLOC_PTR (wp, w, xsize);
  MPN_COPY (PTR(w), xp, xsize);

  ASSERT_MPR_GET_Z (w, m);
}

void
mpr_div__mul (mpr_ptr w, mpr_srcptr x, mpr_srcptr y, mpm_srcptr m)
{
  mp_size_t  msize, xsize, ysize, tsize, talloc, qalloc;
  mp_ptr     tp, wp, qp;
  mp_srcptr  xp, yp;
  mp_limb_t  thigh;
  TMP_DECL (marker);

  ASSERT_MPR_DIV (x, m);
  ASSERT_MPR_DIV (y, m);

  msize = MRSIZE(m);

  xsize = msize;
  xp = RPTR(x);
  MPN_NORMALIZE (xp, xsize);

  ysize = msize;
  yp = RPTR(y);
  MPN_NORMALIZE (yp, ysize);

  if (xsize < ysize)
    MPN_SRCPTR_SWAP (xp,xsize, yp,ysize);

  if (ysize == 0)
    {
      MPN_ZERO (RPTR(w), msize);
      goto done;
    }

  TMP_MARK (marker);

  talloc = xsize + ysize;
  qalloc = msize+1;
  TMP_ALLOC_LIMBS_2 (tp,talloc, qp,qalloc);

  thigh = mpn_mul (tp, xp, xsize, yp, ysize);
  tsize = xsize + ysize;
  if (thigh == 0)
    {
      ASSERT (tsize >= 2);
      thigh = tp[tsize-2];
      tsize--;
    }

  wp = RPTR(w);
  if (tsize < msize)
    {
      MPN_COPY_ZERO_EXTEND (wp,msize, tp,tsize);
    }
  else
    {
      ASSERT (tsize-msize+1 <= qalloc);
      mpn_tdiv_qr (qp, wp, (mp_size_t) 0, tp,tsize, MPTR(m),msize);
    }

  TMP_FREE (marker);

 done:
  ASSERT_MPR_DIV (w, m);
}

/* ENHANCE-ME: Do own loop rather than calling mpz_powm. */
void
mpr_div__pow (mpr_ptr r, mpr_srcptr b, mpz_srcptr e, mpm_srcptr m)
{
  mpz_t      rz, bz;
  mp_srcptr  bp;
  mp_size_t  msize, bsize;

  ASSERT_MPR_DIV (b, m);

  mpz_init (rz);

  bp = RPTR(b);
  msize = MRSIZE(m);
  bsize = msize;
  MPN_NORMALIZE (bp, bsize);
  PTR(bz) = (mp_ptr) bp;
  SIZ(bz) = bsize;
  ASSERT_CODE (ALLOC(bz) = msize);

  mpz_powm (rz, bz, e, MZ(m));

  ASSERT(SIZ(rz) >= 0);
  ASSERT(SIZ(rz) <= msize);
  MPN_COPY_ZERO_EXTEND (RPTR(r),msize, PTR(rz),SIZ(rz));
  mpz_clear (rz);

  ASSERT_MPR_DIV (r, m);
}

void
mpr_div__set_si (mpr_ptr w, long x, mpm_srcptr m)
{
  MPZ_FAKE_SI_VAR (xz, x);
  mpr_mod__set_z (w, xz, m);
}

void
mpr_div__set_ui (mpr_ptr w, unsigned long x, mpm_srcptr m)
{
  mp_size_t  msize;
  mp_ptr     wp;
  mp_limb_t  xlimb = x;

  wp = RPTR(w);
  msize = MRSIZE(m);

#if GMP_NUMB_BITS >= BITS_PER_ULONG
  wp[0] = xlimb;
  MPN_ZERO_EXTEND (wp, (mp_size_t) 1, msize);
#else
  if (UNLIKELY (msize == 2
                && MPTR(m)[1] <= (ULONG_MAX >> GMP_NUMB_BITS)))
    xlimb %= ((MPTR(m)[1] << GMP_NUMB_BITS) | MPTR(m)[0]);
  wp[0] = xlimb & GMP_NUMB_MASK;
  wp[1] = xlimb >> GMP_NUMB_BITS;
  MPN_ZERO_EXTEND (wp, 2, msize);
#endif

  ASSERT_MPR_DIV (w, m);
}

void
mpr_div__sub_ui (mpr_ptr w, mpr_srcptr x, unsigned long y, mpm_srcptr m)
{
  mp_size_t  msize;
  mp_ptr     wp;
  mp_limb_t  c;

  ASSERT_MPR_DIV (x, m);
  ASSERT (MRSIZE(m) >= 2);

  if (UNLIKELY (y > GMP_NUMB_MAX))
    {
      mpr_t  yr;
      mpr_init_set_ui (yr, y, m);
      mpr_sub (w, x, yr, m);
      mpr_clear (yr, m);
      return;
    }

  wp = RPTR(w);
  msize = MRSIZE(m);
  MPN_SUB_1 (c, wp, RPTR(x), msize, (mp_limb_t) y & GMP_NUMB_MASK);

  if (UNLIKELY (c != 0))
    ASSERT_CARRY (mpn_add_n (wp, wp, MPTR(m), msize));

  ASSERT_MPR_DIV (w, m);
}


void
mpm_div__mpm_clear (mpm_ptr m)
{
  __GMP_FREE_FUNC_LIMBS (MNEGONE(m), MRSIZE(m));
  mpm_mod__clear (m);
}

void
mpm_init_div (mpm_ptr m, mpz_srcptr mz)
{
  mp_size_t  msize;
  mp_srcptr  mp;

  msize = SIZ(mz);
  if (msize == 0)
    DIVIDE_BY_ZERO;

  msize = ABS (msize);
  if (msize == 1)
    {
      mpm_init_one (m, PTR(mz)[0]);
      return;
    }

  mpm_mod__init (m, mz);
  msize = MRSIZE(m);
  mp = MPTR(m);

  m->_mpm_clear = mpm_div__mpm_clear;

  m->_mpr_add          = mpr_mod__add;
  m->_mpr_add_ui       = mpr_div__add_ui;
  m->_mpr_addmul_ui    = mpr_mod__addmul_ui;
  m->_mpr_clear        = mpr_gen__clear;
  m->_mpr_div          = mpr_gen__div__invert_mul;
  m->_mpr_div_2exp     = mpr_mod__div_2exp;
  m->_mpr_equal        = mpr_gen__equal;
  m->_mpr_equal_negone = mpr_gen__equal_negone;
  m->_mpr_equal_one    = mpr_div__equal_one;
  m->_mpr_equal_ui     = mpr_div__equal_ui;
  m->_mpr_equal_zero   = mpr_gen__equal_zero;
  m->_mpr_gcd          = mpr_div__gcd;
  m->_mpr_get_ui       = mpr_div__get_ui;
  m->_mpr_get_z        = mpr_div__get_z;
  m->_mpr_invert       = mpr_gen__invert;
  m->_mpr_jacobi       = mpr_gen__jacobi;
  m->_mpr_mul          = mpr_div__mul;
  m->_mpr_mul_2exp     = mpr_mod__mul_2exp;
  m->_mpr_mul_si       = mpr_mod__mul_si;
  m->_mpr_mul_ui       = mpr_mod__mul_ui;
  m->_mpr_neg          = mpr_mod__neg;
  m->_mpr_new          = mpr_gen__new;
  m->_mpr_pow          = mpr_div__pow;
  m->_mpr_set          = mpr_gen__set;
  m->_mpr_set_z        = mpr_mod__set_z;
  m->_mpr_set_si       = mpr_div__set_si;
  m->_mpr_set_ui       = mpr_div__set_ui;
  m->_mpr_sub          = mpr_mod__sub;
  m->_mpr_sub_ui       = mpr_div__sub_ui;
  m->_mpr_submul_ui    = mpr_mod__submul_ui;
  m->_mpr_swap         = mpr_gen__swap;
  m->_mpr_urandomm     = mpr_mod__urandomm;

  {
    mp_ptr     mnegone;
    mp_limb_t  c;
    mnegone = __GMP_ALLOCATE_FUNC_LIMBS (msize);
    MNEGONE(m) = (mpr_ptr) mnegone;
    MPN_SUB_1 (c, mnegone, mp, msize, CNST_LIMB(1));
    ASSERT (c == 0);
    ASSERT_MPR_MOD (m->_mp_negone, m);
  }
}
