/* Shared modular routines, expecting 0<=r<m.

   Copyright 2001, 2002, 2007 Kevin Ryde

   This file is part of MPM.

   MPM is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2, or (at your option) any later version.

   MPM is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
   FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
   details.

   You should have received a copy of the GNU General Public License along
   with MPM; see the file COPYING.  If not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301
   USA.  */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "gmp.h"
#include "gmp-impl.h"
#include "longlong.h"

#include "mpm.h"
#include "mpm-impl.h"



/* change this to "#define TRACE(x) x" for diagnostics */
#define TRACE(x) 


#if GMP_NAIL_BITS == 0
#define MUDIV_QRNND(q, r, nh, nl, m)                                    \
  do {                                                                  \
    ASSERT_LIMB (nh);                                                   \
    ASSERT_LIMB (nl);                                                   \
    udiv_qrnnd_preinv (q, r, nh, nl, MUDIV_DH(m), MUDIV_DH_INV(m));     \
    (r) >>= GMP_NAIL_BITS;                                              \
  } while (0)
#define MUDIV_QRNND_SHIFT  0
#else
#define MUDIV_QRNND(q, r, nh, nl, m)                            \
  do {                                                          \
    ASSERT_LIMB (nh);                                           \
    ASSERT_LIMB (nl);                                           \
    udiv_qrnnd (q, r, nh, (nl) << GMP_NAIL_BITS, MUDIV_DH(m));  \
    (r) >>= GMP_NAIL_BITS;                                      \
  } while (0)
#define MUDIV_QRNND_SHIFT  GMP_NAIL_BITS
#endif


/* Reduce the value w0:{wp,msize}, to the range 0 to m-1.

   Must have w < 2^GMP_NUMB_BITS * m, ie. the biggest multiple which will be
   subtracted is 2^GMP_NUMB_BITS-1.

   This is basically one step of mpn_sb_divrem_mn.  dh and dl are
   pre-normalized, and the high limbs of w are shifted (into n0 etc) are
   shifted a corresponding amount.

   ENHANCE-ME: Make special case routines for msize==1 and 2 so the fetches
   wp[msize-n] can be unconditional.

   ENHANCE-ME: Use udiv_qrnnd_preinv when profitable. */

void
mpr_mod__reduce_1 (mp_limb_t w0, mp_ptr wp, mpm_srcptr m)
{
  mp_size_t  msize;
  mp_srcptr  mp;
  mp_limb_t  q, r, w1, w2, w3, n0, n1, n2, dh, ph, pl, c, mnorm_rshift_mask;
  int        mnorm_lshift, mnorm_rshift;

  TRACE (mpn_trace ("  w0", &w0, (mp_size_t) 1);
         mpn_trace ("   w", wp, MRSIZE(m));
         mpn_trace ("   m", MPTR(m), MRSIZE(m)));

  /* c1:wp < m*2^GMP_NUMB_BITS  */
  ASSERT (w0 < MPTR(m)[MRSIZE(m)-1]
          || (w0 == MPTR(m)[MRSIZE(m)-1]
              && (mpn_cmp (wp+1, MPTR(m), MRSIZE(m)-1) < 0)));

  msize = MRSIZE(m);
  mnorm_lshift = MNORM_LSHIFT(m);
  mnorm_rshift = MNORM_RSHIFT(m);
  mnorm_rshift_mask = MNORM_RSHIFT_MASK(m);

  n0 = w0;
  w1 = wp[msize-1];
  n0 = (w0 << mnorm_lshift)
    | ((w1 >> mnorm_rshift) & mnorm_rshift_mask);
  TRACE (mpn_trace ("    n0", &n0, (mp_size_t) 1));
  ASSERT_LIMB (n0);

  dh = MUDIV_DH(m);
  TRACE (mpn_trace ("    dh", &dh, (mp_size_t) 1));
  if (UNLIKELY (dh == (n0 << GMP_NAIL_BITS)))
    {
      q = GMP_NUMB_MAX;
    }
  else
    {
      w2 = (LIKELY (msize >= 2) ? wp[msize-2] : 0);
      n1 = ((w1 << mnorm_lshift) & GMP_NUMB_MASK)
        | ((w2 >> mnorm_rshift) & mnorm_rshift_mask);
      TRACE (mpn_trace ("    n1", &n1, (mp_size_t) 1));

      udiv_qrnnd (q, r, n0, n1 << GMP_NAIL_BITS, dh);
      r >>= GMP_NAIL_BITS;
      TRACE (mpn_trace ("   q", &q, (mp_size_t) 1);
             mpn_trace ("   r", &r, (mp_size_t) 1));

      /* most cases of q one too big, all cases of two too big */
      umul_ppmm (ph, pl, q, MUDIV_DL(m));
      pl >>= GMP_NAIL_BITS;
      TRACE (mpn_trace ("    ph", &ph, (mp_size_t) 1);
             mpn_trace ("    pl", &pl, (mp_size_t) 1));

      w3 = (LIKELY (msize >= 3) ? wp[msize-3] : 0);
      n2 = ((w2 << mnorm_lshift) & GMP_NUMB_MASK)
        | ((w3 >> mnorm_rshift) & mnorm_rshift_mask);
      TRACE (mpn_trace ("    n2", &n2, (mp_size_t) 1));

      if (CMP_2_GT (ph,pl, r,n2))
        {
          TRACE (printf ("  decrement q=%#lX\n", q));
          q--;
        }
    }

  mp = MPTR(m);
  c = mpn_submul_1 (wp, mp, msize, q);
  TRACE (mpn_trace ("     c", &c, (mp_size_t) 1);
         mpn_trace ("  subm", wp, msize));

  /* addback if q was too big */
  if (UNLIKELY (c != w0))
    {
      ASSERT (c == w0 + 1);
      ASSERT_CARRY (mpn_add_n (wp, wp, mp, msize));
    }

  ASSERT (mpn_cmp (wp, mp, msize) < 0);
}


void
mpr_mod__add (mpr_ptr w, mpr_srcptr x, mpr_srcptr y, mpm_srcptr m)
{
  mp_size_t  msize;
  mp_ptr     wp;
  mp_srcptr  mp;

  ASSERT_MPR_MOD (x, m);
  ASSERT_MPR_MOD (y, m);
  TRACE (printf ("mpr_redc__add\n");
         mpn_trace ("  x", RPTR(x), MRSIZE(m));
         mpn_trace ("  y", RPTR(y), MRSIZE(m)));

  mp = MPTR(m);
  wp = RPTR (w);
  msize = MRSIZE(m);

  if (mpn_add_n (wp, RPTR(x), RPTR(y), msize))
    goto subtract;
  MPN_CMP_LT_ACTION (wp, mp, msize, goto done);
 subtract:
  mpn_sub_n (wp, wp, mp, msize);

 done:
  ASSERT_MPR_MOD (w, m);
}

void
mpr_mod__addmul_ui (mpr_ptr w, mpr_srcptr x, unsigned long y, mpm_srcptr m)
{
  mp_size_t  msize;
  mp_ptr     wp;
  mp_limb_t  c;

  ASSERT_MPR_MOD (w, m);
  ASSERT_MPR_MOD (x, m);

  if (UNLIKELY (y > GMP_NUMB_MAX))
    {
      mpr_t  yr;
      TMP_DECL (marker);
      TMP_MARK (marker);
      MPR_TMP_INIT (yr, m);
      mpr_div__set_ui (yr, y, m);
      mpr_div__mul (yr, x, yr, m);
      mpr_mod__add (w, w, yr, m);
      TMP_FREE (marker);
      return;
    }

  wp = RPTR(w);
  msize = MRSIZE(m);
  c = mpn_addmul_1 (wp, RPTR(x), msize, (mp_limb_t) y);
  mpr_mod__reduce_1 (c, wp, m);

  ASSERT_MPR_MOD (w, m);
}

/* ENHANCE-ME: Tune how long bit-by-bit is faster than mpn_submul_1.  Will
   this depend on msize?

   ENHANCE-ME: For big e do a powering algorithm to get a factor 2^-e to
   multiply.  Tune for what sizes to do this.  */

void
mpr_mod__div_2exp (mpr_ptr w, mpr_srcptr x, unsigned long e, mpm_srcptr m)
{
  mp_ptr     wp;
  mp_srcptr  xp;
  mp_size_t  msize;
  unsigned long  i;

  ASSERT_MPR_MOD (x, m);

  if (UNLIKELY ((MPTR(m)[0] & 1) == 0))
    DIVIDE_BY_ZERO;

  wp = RPTR(w);
  xp = RPTR(x);
  msize = MRSIZE(m);

  if (e == 0)
    {
      if (wp != xp)
        MPN_COPY (wp, xp, msize);
      return;
    }

  //  if (BELOW_THRESHOLD_ULONG (e, MPR_MOD__DIV_2EXP_INV_THRESHOLD)
  {
    mp_srcptr  dp = MDIV2CEIL(m);
    for (i = 0; i < e; i++)
      {
        if (mpn_rshift (wp, xp, msize, 1) != 0)
          ASSERT_NOCARRY (mpn_add_n (wp, wp, dp, msize));
        xp = wp;
      }
    return;
  }

  MPN_COPY (wp, xp, msize);
  for (;;)
    {
      mp_limb_t  q, qmask, c;
      mp_srcptr  mp = MPTR(m);
      qmask = (CNST_LIMB(1) << MIN (e, GMP_NUMB_BITS)) - 1;
      q = (wp[0] * MSMALLINV(m)[0]) & qmask;
      c = mpn_submul_1 (wp, mp, msize, q);
      MPN_COPY_DECR (wp, wp+1, msize-1);
      wp[msize-1] = -c;
      if (c != 0)
        ASSERT_CARRY (mpn_add_n (wp, wp, mp, msize));
    }
  abort ();
}

/* ENHANCE-ME: Do a powering algorithm for 2^y when y is big. */
void
mpr_mod__mul_2exp (mpr_ptr w, mpr_srcptr x, unsigned long y, mpm_srcptr m)
{
  mp_size_t  msize, psize, ylimbs, palloc, qalloc;
  mp_ptr     pp, wp, qp;
  mp_srcptr  xp;
  int        yshift;
  mp_limb_t  c;
  TMP_DECL (marker);

  ASSERT_MPR_MOD (x, m);

  TMP_MARK (marker);
  msize = MRSIZE(m);

  ylimbs = y / GMP_NUMB_BITS;

  palloc = msize+ylimbs+1;
  qalloc = ylimbs+2;
  TMP_ALLOC_LIMBS_2 (pp,palloc, qp,qalloc);

  MPN_ZERO (pp, ylimbs);
  psize = ylimbs + msize;
  xp = RPTR(x);
  yshift = y % GMP_NUMB_BITS;
  if (yshift == 0)
    MPN_COPY (pp+ylimbs, xp, msize);
  else
    {
      c = mpn_lshift (pp+ylimbs, xp, msize, yshift);
      pp[psize] = c;
      psize += (c != 0);
    }
  ASSERT (psize <= palloc);

  wp = RPTR(w);
  ASSERT (qalloc >= psize-msize+1);
  mpn_tdiv_qr (qp, wp, (mp_size_t) 0, pp,psize, MPTR(m),msize);
  TMP_FREE (marker);

  ASSERT_MPR_MOD (w, m);
}

/* ENHANCE-ME: Maybe inline the neg here. */
void
mpr_mod__mul_si (mpr_ptr w, mpr_srcptr x, long y, mpm_srcptr m)
{
  if (y < 0)
    {
      mpr_mod__neg (w, x, m);
      x = w;
    }
  mpr_mod__mul_ui (w, x, ABS(y), m);
}

void
mpr_mod__mul_ui (mpr_ptr w, mpr_srcptr x, unsigned long y, mpm_srcptr m)
{
  mp_size_t  msize;
  mp_ptr     wp;
  mp_limb_t  c;

  TRACE (printf ("mpr_mod__mul_ui\n");
         mpn_trace ("   x", RPTR(x), MRSIZE(m));
         printf    ("   y=%#lX\n", y);
         mpn_trace ("   m", MPTR(m), MRSIZE(m)));

  ASSERT_MPR_MOD (x, m);

  if (UNLIKELY (y > GMP_NUMB_MAX))
    {
      mpr_t  yr;
      TMP_DECL (marker);
      TMP_MARK (marker);
      MPR_TMP_INIT (yr, m);
      mpr_div__set_ui (yr, y, m);
      mpr_div__mul (w, x, yr, m);
      TMP_FREE (marker);
      return;
    }

  wp = RPTR(w);
  msize = MRSIZE(m);
  c = mpn_mul_1 (wp, RPTR(x), msize, (mp_limb_t) y);
  mpr_mod__reduce_1 (c, wp, m);

  ASSERT_MPR_MOD (w, m);
}

void
mpr_mod__neg (mpr_ptr w, mpr_srcptr x, mpm_srcptr m)
{
  mp_size_t  i, msize;
  mp_srcptr  xp;

  ASSERT_MPR_MOD (x, m);

  msize = MRSIZE(m);
  xp = RPTR(x);
  i = 0;
  do {
    if (LIKELY (xp[i] != 0))
      {
        ASSERT_NOCARRY (mpn_sub_n (RPTR(w), MPTR(m), xp, msize));
        return;
      }
    i++;
  } while (i < msize);

  if (w != x)
    MPN_ZERO (RPTR(w), msize);
}

void
mpr_mod__set_z (mpr_ptr w, mpz_srcptr x, mpm_srcptr m)
{
  mp_size_t  msize, xsize, abs_xsize, qsize;
  mp_srcptr  mp, xp;
  mp_ptr     wp, qp;
  TMP_DECL (marker);

  msize = MRSIZE(m);
  xsize = SIZ(x);
  wp = RPTR(w);
  xp = PTR(x);
  mp = MPTR(m);
  abs_xsize = ABS (xsize);

  if (LIKELY (abs_xsize < msize))
    {
      if (xsize < 0)
        {
          mp_limb_t  c;
          MPN_SUB (c, wp, MPTR(m), msize, xp, abs_xsize);
          ASSERT (c == 0);
        }
      else
        MPN_COPY_ZERO_EXTEND (wp, msize, xp, xsize);
      return;
    }

  TMP_MARK (marker);
  qsize = abs_xsize - msize + 1;
  qp = TMP_ALLOC_LIMBS (qsize);

  mpn_tdiv_qr (qp, wp, (mp_size_t) 0, xp,abs_xsize, mp,msize);
  TMP_FREE (marker);

  if (xsize < 0)
    mpr_mod__neg (w, w, m);

  ASSERT_MPR_MOD (w, m);
}

void
mpr_mod__sub (mpr_ptr w, mpr_srcptr x, mpr_srcptr y, mpm_srcptr m)
{
  mp_size_t  msize;
  mp_ptr     wp;

  ASSERT_MPR_MOD (x, m);
  ASSERT_MPR_MOD (y, m);

  msize = MRSIZE(m);
  wp = RPTR (w);
  if (mpn_sub_n (wp, RPTR(x), RPTR(y), msize))
    ASSERT_CARRY (mpn_add_n (wp, wp, MPTR(m), msize));

  ASSERT_MPR_MOD (w, m);
}

void
mpr_mod__submul_ui (mpr_ptr w, mpr_srcptr x, unsigned long y, mpm_srcptr m)
{
  mp_ptr     wp;
  mp_srcptr  mp;
  mp_size_t  msize;
  mp_limb_t  c, c2;

  ASSERT_MPR_MOD (w, m);
  ASSERT_MPR_MOD (x, m);
  ASSERT (MRSIZE(m) >= 1);

  if (UNLIKELY (y > GMP_NUMB_MAX))
    {
      mpr_t  yr;
      TMP_DECL (marker);
      TMP_MARK (marker);
      MPR_TMP_INIT (yr, m);
      mpr_div__set_ui (yr, y, m);
      mpr_div__mul (yr, x, yr, m);
      mpr_mod__sub (w, w, yr, m);
      TMP_FREE (marker);
      return;
    }

  msize = MRSIZE(m);
  wp = RPTR(w);
  c = mpn_submul_1 (wp, RPTR(x), msize, (mp_limb_t) y);
  if (UNLIKELY (c == 0))
    return;

  mp = MPTR(m);
  c2 = mpn_add_n (wp+1, wp+1, mp, msize-1);
  c = mp[msize-1] + c2 - c;
  mpr_mod__reduce_1 (c, wp, m);

  ASSERT_MPR_MOD (w, m);
}

void
mpr_mod__urandomm (mpr_ptr w, gmp_randstate_ptr rstate, mpm_srcptr m)
{
  mp_ptr     wp;
  mp_srcptr  mp;
  int        i;
  mp_size_t  msize;
  unsigned long  mbits;

  wp = RPTR(w);
  msize = MRSIZE(m);
  mp = MPTR(m);
  mbits = MBITS(m);

  i = 0;
  for (;;)
    {
      _gmp_rand (wp, rstate, mbits);
      MPN_CMP_LT_ACTION (wp, mp, msize, goto done);

      /* in case _gmp_rand returns all 1 bits */
      i++;
      if (i >= 100)
        {
          MPN_ZERO (wp, msize);
          break;
        }
    }

 done:
  ASSERT_MPR_MOD (w, m);
}


void
mpm_mod__clear (mpm_ptr m)
{
  mp_size_t  msize;
  msize = MRSIZE(m);
  __GMP_FREE_FUNC_LIMBS (MNORM(m), msize);
  __GMP_FREE_FUNC_LIMBS (MDIV2CEIL(m), msize);
}

void
mpm_mod__init (mpm_ptr m, mpz_srcptr mz)
{
  mp_size_t  msize;
  mp_ptr     mp;
  int        mnorm_lshift;

  if (mpz_cmp_ui (mz, 1L) <= 0)
    DIVIDE_BY_ZERO;

  ASSERT_CODE (memset (m, '\0', sizeof (mpm_t)));

  msize = SIZ (mz);
  if (msize == 0)
    DIVIDE_BY_ZERO;

  /* m */
  msize = ABS (msize);
  ALLOC(MZ(m)) = msize;
  SIZ(MZ(m)) = msize;
  mp = __GMP_ALLOCATE_FUNC_LIMBS (msize);
  MPTR(m) = mp;
  MPN_COPY (mp, PTR(mz), msize);

  MRSIZE(m) = msize;
  MRALLOC(m) = msize;

  /* mnorm */
  {
    mp_ptr     mnorm;
    mp_limb_t  h, l;

    count_leading_zeros (mnorm_lshift, mp[msize-1]);
    mnorm_lshift -= GMP_NAIL_BITS;
    MNORM_LSHIFT(m) = mnorm_lshift;
    MNORM_RSHIFT(m) = (GMP_NUMB_BITS - mnorm_lshift);
    MNORM_RSHIFT_MASK(m) = (mnorm_lshift == 0 ? 0 : GMP_NUMB_MASK);

    mnorm = __GMP_ALLOCATE_FUNC_LIMBS (msize);
    MNORM(m) = mnorm;
    if (mnorm_lshift == 0)
      MPN_COPY (mnorm, mp, msize);
    else
      ASSERT_NOCARRY (mpn_lshift (mnorm, mp, msize, mnorm_lshift));
    TRACE (mpn_trace ("  mnorm", mnorm, msize));

    h = mnorm[msize-1];
    l = (LIKELY (msize >= 2) ? mnorm[msize-2] : 0);
    MUDIV_DH(m) = h << GMP_NAIL_BITS;
    MUDIV_DL(m) = l << GMP_NAIL_BITS;
    if (GMP_NAIL_BITS == 0)
      invert_limb (MUDIV_DH_INV(m), h);
  }

  /* mbits */
  MBITS(m) = msize * GMP_NUMB_BITS - mnorm_lshift;

  /* mdiv2ceil */
  {
    mp_ptr  mdiv2ceil;
    mdiv2ceil = __GMP_ALLOCATE_FUNC_LIMBS (msize);
    MDIV2CEIL(m) = mdiv2ceil;
    if (mpn_rshift (mdiv2ceil, mp, msize, 1))
      MPN_INCR_U (mdiv2ceil, msize, CNST_LIMB(1));
  }
}
