/* mpn_modlimbs_invert -- multiplicative inverse mod 2^(n*BITS_PER_MP_LIMB).

   Copyright 2001, 2002, 2007 Kevin Ryde

   This file is part of MPM.

   MPM is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2, or (at your option) any later version.

   MPM is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
   FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
   details.

   You should have received a copy of the GNU General Public License along
   with MPM; see the file COPYING.  If not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301
   USA.  */

#include "gmp.h"
#include "gmp-impl.h"
#include "longlong.h"

#include "mpm.h"
#include "mpm-impl.h"


/* ENHANCE-ME: This is not a particularly good implementation.  Should use
   the fact that the low half of inv*src is 1 at each step and the low half
   of the inverse is unchanged at each step.  */


/* Change this to "#define TRACE(x) x" for some traces. */
#define TRACE(x) 


#if WANT_ASSERT
#define ASSERT_INVERSE_2(ih,il, nh,nl)          \
  do {                                          \
    mp_limb_t  __ph, __pl;                      \
    umul_p2m2m2 (__ph, __pl, ih,il, nh, nl);    \
    ASSERT (__pl == 1);                         \
    ASSERT (__ph == 0);                         \
  } while (0)
#else
#define ASSERT_INVERSE_2()
#endif

#ifndef mpn_mullow_n
#define mpn_mullow_n(wp,xp,yp,size)   mpn_mul_n (wp, xp, yp, size)
#endif

#if WANT_ASSERT
#define ASSERT_INVERSE()                                \
  do {                                                  \
    mp_size_t  j;                                       \
    int        ok;                                      \
    ASSERT (qalloc >= 2*dsize);                         \
    mpn_mullow_n (q, dst, src, dsize);                  \
    ok = (q[0] == 1);                                   \
    for (j = 1; j < dsize && ok; j++)                   \
      if (q[j] != 0)                                    \
        ok = 0;                                         \
    if (!ok)                                            \
      TRACE (mpn_trace ("  wrong prod", q, dsize));     \
    ASSERT (ok);                                        \
  } while (0)
#else
#define ASSERT_INVERSE()  do { } while (0)
#endif



/* Set {d,n} to satisfy {d,n} * {s,n} == 1 modulo 2^(n*BITS_PER_MP_LIMB).  */
void
mpn_modlimbs_invert (mp_ptr dst, mp_srcptr src, mp_size_t size)
{
  mp_ptr     p, q;
  mp_limb_t  inv1, inv0, src1, src0, c, dummy;
  mp_size_t  palloc, qalloc, qsize, dsize, step_dsize;
  unsigned   cnt;
  int        i;
  TMP_DECL (marker);

  TRACE (printf ("mpn_modlimbs_invert size=%ld\n", size);
         mpn_trace ("  src", src, size));

  ASSERT (size >= 1);
  ASSERT (! MPN_OVERLAP_P (dst,size, src,size));
  ASSERT ((src[0] & 1) != 0);
  ASSERT_MPN (src, size);

  src0 = src[0];
  modlimb_invert (inv0, src0);
  dst[0] = inv0;
  if (size == 1)
    return;

  src1 = src[1];
  modlimb_invert_1to2 (inv1, inv0, src1, src0);
  dst[1] = inv1;
  if (size == 2)
    return;

  TMP_MARK (marker);
  palloc = size+1;
  qalloc = 2*(size+1);
  TMP_ALLOC_LIMBS_2 (p,palloc, q,qalloc);

  /* i = ceil(log2(size)) - 2 */
  count_leading_zeros (cnt, (mp_limb_t) size - 1);
  i = BITS_PER_MP_LIMB - cnt - 2;
  ASSERT ((1L << (i+2)) >= size);
  ASSERT (i >= 0);

  dst[0] = inv0;
  dst[1] = inv1;
  dsize = 2;

  do
    {
      TRACE (printf    ("  i=%d  dsize=%ld\n", i, dsize);
             mpn_trace ("  dst", dst, dsize));
      ASSERT_MPN (dst, dsize);
      ASSERT_INVERSE ();

      /* inv*inv */
      ASSERT (palloc >= 2*dsize);
      mpn_sqr_n (p, dst, dsize);

      /* inv*inv*src */
      ASSERT (qalloc >= 4*dsize);
      qsize = MIN (2*dsize, size);
      mpn_mullow_n (q, p, src, qsize);

      /* 2*inv */
      ASSERT (palloc >= dsize+1);
      c = mpn_lshift (dst, dst, dsize, 1);

      /* 2*inv-inv*inv*src */
      c += 1-mpn_sub_n (dst, dst, q, dsize);
      step_dsize = dsize - (((2*dsize-1) << i) >= size);
      ASSERT (step_dsize >= 1);
      ASSERT (dsize+step_dsize <= size);
      mpn_com_n (dst+dsize, q+dsize, step_dsize);
      MPN_ADD_1 (dummy, dst+dsize, dst+dsize, step_dsize, c);

      dsize += step_dsize;
    }
  while (--i >= 0);

  TRACE (printf    ("  i=%d  dsize=%ld\n", i, dsize);
         mpn_trace ("  dst", dst, dsize));
  ASSERT (dsize == size);
  ASSERT_INVERSE ();

  TMP_FREE (marker);
}
